#ifndef COLOR_H__
#define COLOR_H__

#define PX_WIDTH  (13)
#define PX_HEIGHT (13)
#define NUM_PIXELS (PX_WIDTH*PX_HEIGHT)

#define GRID(x,y)        ((y)*PX_WIDTH+(x))
#define LERP(a,b,p)      ( (a) + ( ((b)-(a)) * p ) )
#define randf()          (random(0xffffff) * (1.0/0xffffff))
#define randRange(a,b)   LERP( a, b, randf() )

// RGB: Range of each channel is 0.0...1.0
struct RGB {
  float r;  // red
  float g;  // green 
  float b;  // blue
};

// Interpolate RGB values to create a new RGB color.
// Progress: 0==100% c0, 1==100% c1.
RGB lerpRGB( RGB c0, RGB c1, float progress ) {
  return (RGB){ LERP( c0.r, c1.r, progress ), LERP( c0.g, c1.g, progress ), LERP( c0.b, c1.b, progress ) };
}

// Ranges: 0.0...1.0
struct HSV {
  float h;  // hue. Range: 0..1
  float s;  // saturation
  float v;  // value (brightness)
};

RGB hsv2rgb( float hF, float sF, float vF ) {
  // Expand H by 6 times for the 6 color regions. Wrap the remainder to 0..1
  float hWrap = (hF*6.0);
  hWrap = hWrap - floor(hWrap);  // floating point remainder. 0..1 for each of the 6 color regions
  
  float v = vF;  // top (max) value
  float p = vF * (1.0-sF);  // bottom (min) value
  float q = vF * (1.0-(hWrap*sF));  // falling (yellow->green: the red channel falls)
  float t = vF * (1.0 - ((1.0-hWrap)*sF));  // rising (red->yellow: the green channel rises)
  
  // Need to find the correct color region that the hue belongs to.
  // Hue can have a negative value, so need to step it to positive space, so the modulus % operator behaves as expected.
  float hF_pos = hF;
  if( hF_pos < 0.0 ) {
    hF_pos += ceil(-hF_pos);
  }
  uint8_t hue_i = (uint8_t)(floor( hF_pos * 6.0 )) % 6;
  
  switch( hue_i ) {
    case 0:  return (RGB){v,t,p};  // red -> yellow
    case 1:  return (RGB){q,v,p};  // yellow -> green
    case 2:  return (RGB){p,v,t};  // green -> cyan
    case 3:  return (RGB){p,q,v};  // cyan -> blue
    case 4:  return (RGB){t,p,v};  // blue -> magenta
    case 5:  return (RGB){v,p,q};  // magenta -> red
  }
  
  return (RGB){0,0,0};  // sanity escape
};

HSV rgb2hsv( float inR, float inG, float inB ) {
  float outH, outS, outV;
  float delta;

  float _min = min( inR, min( inG, inB ));
  float _max = max( inR, max( inG, inB ));

  delta = _max - _min;
  
  // Grayscale / black / white?
  if( delta==0.0 ) {
    return (HSV){ _max, _max, _max };
  }

  outS = (delta / _max);               // Saturation
  outV = _max;                           // Value
  
  if( inR == _max ) {                       
    outH = ( inG - inB ) / delta;        // between yellow & magenta
  } else if( inG >= _max ) {
    outH = 2.0 + ( inB - inR ) / delta;  // between cyan & yellow
  } else {
    outH = 4.0 + ( inR - inG ) / delta;  // between magenta & cyan
  }

  outH *= (1.0/6.0);  // range: 0..1

  if( outH < 0.0 ) {
    outH += 1.0;
  }

  return (HSV){ outH, outS, outV };
};

PROGMEM prog_uchar GAMMA[] = {
  0,   1,   1,   1,   1,   1,   1,   1,   1,   1,   1,   1,   1,   1,   1,   1,
  1,   1,   1,   1,   1,   1,   1,   1,   1,   1,   1,   1,   1,   1,   1,   1,
  1,   2,   2,   2,   2,   2,   2,   2,   2,   3,   3,   3,   3,   3,   4,   4,
  4,   4,   4,   5,   5,   5,   5,   6,   6,   6,   6,   7,   7,   7,   7,   8,
  8,   8,   9,   9,   9,  10,  10,  10,  11,  11,  12,  12,  12,  13,  13,  14,
 14,  15,  15,  15,  16,  16,  17,  17,  18,  18,  19,  19,  20,  20,  21,  22,
 22,  23,  23,  24,  25,  25,  26,  26,  27,  28,  28,  29,  30,  30,  31,  32,
 33,  33,  34,  35,  36,  36,  37,  38,  39,  40,  40,  41,  42,  43,  44,  45,
 46,  46,  47,  48,  49,  50,  51,  52,  53,  54,  55,  56,  57,  58,  59,  60,
 61,  62,  63,  64,  65,  67,  68,  69,  70,  71,  72,  73,  75,  76,  77,  78,
 80,  81,  82,  83,  85,  86,  87,  89,  90,  91,  93,  94,  95,  97,  98,  99,
101, 102, 104, 105, 107, 108, 110, 111, 113, 114, 116, 117, 119, 121, 122, 124,
125, 127, 129, 130, 132, 134, 135, 137, 139, 141, 142, 144, 146, 148, 150, 151,
153, 155, 157, 159, 161, 163, 165, 166, 168, 170, 172, 174, 176, 178, 180, 182,
184, 186, 189, 191, 193, 195, 197, 199, 201, 204, 206, 208, 210, 212, 215, 217,
219, 221, 224, 226, 228, 231, 233, 235, 238, 240, 243, 245, 248, 250, 253, 255 
};

uint32_t rgb2web( RGB rgb, boolean doGamma ){
  if( doGamma ) {
    return (pgm_read_byte(&GAMMA[(uint8_t)(rgb.r*0xff)]) << 16) |
           (pgm_read_byte(&GAMMA[(uint8_t)(rgb.g*0xff)]) <<  8) |
           (pgm_read_byte(&GAMMA[(uint8_t)(rgb.b*0xff)]));
  }
  
  return ((uint8_t)(rgb.r*0xff)<<16) | ((uint8_t)(rgb.g*0xff)<<8) | (uint8_t)(rgb.b*0xff);
}

#endif


